"""
TotalCloud API module

This module contains ways to interact with the Qualys TotalCloud APIs. 
"""

from .get_connectors import (
    get_connectors,
    get_connector_details,
    get_aws_base_account,
)

from .get_metadata import get_control_metadata
from .evaluations import (
    get_evaluation,
    get_account_evaluation,
    get_resources_evaluated_by_control,
)
from .get_inventory import get_inventory
from .get_resource_details import get_resource_details
from .remediation_log import get_remediation_activities



"""
Contains the PostmanCollection dataclass
"""

from dataclasses import dataclass, asdict
from typing import Union


@dataclass
class PostmanCollection:
    """
    Represents a Postman Collection in WAS
    """

    id: Union[str, int] = None
    name: str = None
    content: str = None
    fileSize: Union[str, int] = None

    def __post_init__(self):
        if self.id and isinstance(self.id, str):
            try:
                self.id = int(self.id)
            except ValueError:
                raise ValueError("id must be numeric")

        if self.fileSize and isinstance(self.fileSize, str):
            try:
                self.fileSize = int(self.fileSize)
            except ValueError:
                raise ValueError("fileSize must be numeric")

    def __str__(self) -> str:
        return self.name

    def __int__(self) -> int:
        return self.id

    def size(self) -> int:
        return self.fileSize

    def to_dict(self):
        """
        Converts the object to a dictionary
        """
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        Creates a PostmanCollection object from a dictionary
        """
        return cls(**data)



"""
Contains the dataclasses for a Qualys AWS Connector 
and subclasses for AWS/Azure connectors.
"""

from dataclasses import dataclass, asdict
from typing import Optional, Union
from datetime import datetime

from ...base.base_list import BaseList


@dataclass
class Connector:
    """
    Connector - represents a Qualys Connector resource record.

    NOTE: You should not use this class directly. Use one of the
    AWS/Azure-specific subclasses.
    """

    name: Optional[str] = None
    connectorId: Optional[str] = None
    description: Optional[str] = None
    provider: Optional[str] = None
    state: Optional[str] = None
    totalAssets: Optional[int] = None
    lastSyncedOn: Optional[Union[str, datetime]] = None
    nextSyncedOn: Optional[str] = None
    remediationEnabled: Optional[bool] = None
    qualysTags: Optional[BaseList[str]] = None
    isGovCloud: Optional[bool] = None
    isChinaRegion: Optional[bool] = None
    accountAlias: Optional[str] = None
    isDisabled: Optional[bool] = None
    pollingFrequency: Optional[Union[dict, str]] = None
    error: Optional[str] = None
    externalId: Optional[str] = None
    portalConnectorUuid: Optional[str] = None
    isPortalConnector: Optional[bool] = None
    groups: Optional[BaseList[str]] = None

    def __post_init__(self):
        """
        __post_init__ - post initialization method for the Connector dataclass
        """
        if self.pollingFrequency:
            s = f"{self.pollingFrequency.get('hours', 0)}h, {self.pollingFrequency.get('minutes', 0)}m"
            setattr(self, "pollingFrequency", s)

        if self.qualysTags:
            bl = BaseList()
            data = self.qualysTags
            if isinstance(data, dict):
                data = [data]
            for tag in data:
                bl.append(tag["tagName"])
            setattr(self, "qualysTags", bl)

        DT_FIELDS = ["lastSyncedOn", "nextSyncedOn"]

        for field in DT_FIELDS:
            if getattr(self, field):
                if isinstance(getattr(self, field), datetime):
                    continue
                else:
                    setattr(
                        self,
                        field,
                        datetime.strptime(
                            getattr(self, field), "%a %b %d %H:%M:%S %Z %Y"
                        ),
                    )

        if self.groups:
            bl = BaseList()
            data = self.groups
            if isinstance(data, dict):
                data = [data]
            for group in data:
                bl.append(group["name"])
            setattr(self, "groups", bl)

    def to_dict(self):
        """
        to_dict - returns the Connector object as a dictionary
        """
        return asdict(self)

    def __int__(self):
        return self.connectorId

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    def __getitem__(self, key):
        return self.to_dict()[key]

    def __setitem__(self, key, value):
        setattr(self, key, value)

    @classmethod
    def from_dict(cls, data):
        """
        from_dict - creates a Connector object from a dictionary
        """
        return cls(**data)


@dataclass
class AWSConnector(Connector):
    """
    AWS-specific subclass of the Connector dataclass

    Adds awsAccountId and baseAccountId fields to the Connector dataclass
    """

    awsAccountId: Optional[Union[str, int]] = None
    baseAccountId: Optional[str] = None
    arn: Optional[str] = None

    def __post_init__(self):
        INT_FIELDS = ["awsAccountId", "baseAccountId"]
        for field in INT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), int):
                setattr(self, field, int(getattr(self, field)))

        super().__post_init__()


@dataclass
class AzureConnector(Connector):
    """
    Azure-specific subclass of the Connector dataclass
    """

    subscriptionId: Optional[str] = None
    tenantId: Optional[str] = None
    applicationId: Optional[str] = None
    subscriptionName: Optional[str] = None
    directoryId: Optional[str] = None

    def __post_init__(self):
        super().__post_init__()



"""
Contains the datatype filter mappings for
operators
"""

ENDPOINT_MAPPINGS = {
    "count_webapps": {
        "id": "INTEGER",
        "name": "TEXT",
        "url": "TEXT",
        "tags_name": "TEXT",
        "tags_id": "INTEGER",
        "createdDate": "DATE",
        "updatedDate": "DATE",
        "isScheduled": "BOOLEAN",
        "isScanned": "BOOLEAN",
        "lastScan_status": "KEYWORD",
        "lastScan_date": "DATE",
    },
    "get_webapps": {
        "id": "INTEGER",
        "name": "TEXT",
        "url": "TEXT",
        "tags_name": "TEXT",
        "tags_id": "INTEGER",
        "createdDate": "DATE",
        "updatedDate": "DATE",
        "isScheduled": "BOOLEAN",
        "isScanned": "BOOLEAN",
        "lastScan_status": "KEYWORD",
        "lastScan_date": "DATE",
        "verbose": "BOOLEAN",
    },
    "get_webapp_details": {},
    "create_webapp": {
        "name": "TEXT",
        "url": "TEXT",
        "authRecord_id": "INTEGER",
        "uris": "TEXT",
        "tag_ids": "TEXT",
        "domains": "TEXT",
        "scannerTag_ids": "TEXT",
    },
    "delete_webapp": {
        "id": "INTEGER",
        "name": "TEXT",
        "url": "TEXT",
        "tags_name": "TEXT",
        "tags_id": "INTEGER",
        "createdDate": "DATE",
        "updatedDate": "DATE",
        "isScheduled": "BOOLEAN",
        "isScanned": "BOOLEAN",
        "lastScan_status": "KEYWORD",
        "lastScan_date": "DATE",
        "removeFromSubscription": "BOOLEAN",
    },
    "get_selenium_script": {
        "id": "INTEGER",
        "crawlingScripts_id": "INTEGER",
    },
    "count_authentication_records": {
        "id": "INTEGER",
        "name": "TEXT",
        "tags": "INTEGER",
        "tags_name": "TEXT",
        "tags_id": "INTEGER",
        "createdDate": "DATE",
        "updatedDate": "DATE",
        "lastScan_date": "DATE",
        "lastScan_authStatus": "KEYWORD",
        "isUsed": "BOOLEAN",
        "contents": "KEYWORD",
    },
}

# Build update_webapp with create_webapp as a base:
ENDPOINT_MAPPINGS["update_webapp"] = ENDPOINT_MAPPINGS["create_webapp"]
ENDPOINT_MAPPINGS["update_webapp"]["webappId"] = "INTEGER"
ENDPOINT_MAPPINGS["update_webapp"]["attributes"] = "TEXT"
ENDPOINT_MAPPINGS["update_webapp"]["urlExcludelist"] = "TEXT"
ENDPOINT_MAPPINGS["update_webapp"]["urlAllowlist"] = "TEXT"
ENDPOINT_MAPPINGS["update_webapp"]["postDataExcludelist"] = "TEXT"
ENDPOINT_MAPPINGS["update_webapp"]["useSitemap"] = "BOOLEAN"
ENDPOINT_MAPPINGS["update_webapp"]["headers"] = "TEXT"

# get_authentication_records is the same as count_authentication_records:
ENDPOINT_MAPPINGS["get_authentication_records"] = ENDPOINT_MAPPINGS[
    "count_authentication_records"
]

FILTER_MAPPING = {
    "INTEGER": ["EQUALS", "NOT EQUALS", "GREATER", "LESSER", "IN"],
    "TEXT": ["CONTAINS", "EQUALS", "NOT EQUALS"],
    "DATE": ["EQUALS", "NOT EQUALS", "GREATER", "LESSER"],
    "KEYWORD": [
        "EQUALS",
        "NOT EQUALS",
        "IN",
        "NONE",
        "NOT_USED",
        "SUCCESSFUL",
        "FAILED",
        "PARTIAL",
        "FORM_STANDARD",
        "FORM_CUSTOM",
        "FORM_SELENIUM",
        "SERVER_BASIC",
        "SERVER_DIGEST",
        "SERVER_NTLM",
        "CERTIFICATE",
        "OAUTH2_AUTH_CODE",
        "OAUTH2_IMPLICIT",
        "OAUTH2_PASSWORD",
        "OAUTH2_CLIENT_CREDS",
    ],
    "BOOLEAN": ["EQUALS", "NOT EQUALS"],
}



"""
Contains the WebAppAuthRecord class for the WAS module.
"""

from dataclasses import dataclass, asdict
from datetime import datetime
from operator import ge
from typing import Union, Literal

from requests import get

from qualysdk.was.data_classes.Tag import WASTag

from ...base.base_list import BaseList
from .Comment import Comment
from .WebAppAuthFormRecord import WebAppAuthFormRecord
from .WebAppAuthServerRecord import WebAppAuthServerRecord


def build_oauth2_record(dataclass) -> None:
    """
    Builds the OAuth2 record for a WebAppAuthRecord

    Args:
        dataclass: The dataclass to build the OAuth2 record for

    Returns:
        None
    """

    for field in [
        "grantType",
        "accessTokenUrl",
        "clientId",
        "clientSecret",
        "scope",
        "seleniumCreds",
    ]:
        setattr(
            dataclass,
            f"oauth2Record_{field}",
            getattr(dataclass, "oauth2Record").get(field),
        )

    setattr(
        dataclass,
        "oauth2Record_seleniumCreds",
        getattr(dataclass, "oauth2Record").get("seleniumCreds") == "true",
    )

    setattr(dataclass, "oauth2Record", None)


def handle_qualys_list(data: dict, key: str) -> tuple[int, BaseList]:
    """
    Normalizes list responses from Qualys

    Args:
        data (dict): The data to normalize
        key (str): The key to normalize

    Returns:
        tuple[int, BaseList]: The count and the list
    """

    # Mappings for object creation:
    MAPPINGS = {
        "WebAppAuthFormRecordField": WebAppAuthFormRecord,
        "WebAppAuthServerRecordField": WebAppAuthServerRecord,
        "Comment": Comment,
        "Tag": WASTag,
        "formRecord": WebAppAuthFormRecord,
        "serverRecord": WebAppAuthServerRecord,
    }

    # Easy short circuit, with walrus to save value:
    if data.get("count") == "0":
        return 0, BaseList()

    list_key = data.get("list")

    fields_record = False
    # Check for the record-specific fields key:
    if not list_key and len(data.keys()) > 1:
        list_key = data.get("fields", {}).get("list")
        # Raise the data up by one key:
        if not list_key:
            return int(data["fields"].get("count", 0)), BaseList()

        to_raise = list(list_key.keys())[0]
        list_key = list_key[to_raise]
        fields_record = True

    if not list_key:
        return int(data.get("count", 0)), BaseList()

    if not fields_record:
        list_data = list_key.get(key)
    else:
        list_data = list_key

    if isinstance(list_data, dict):
        list_data = [list_data]

    base_list = BaseList()

    for i in list_data:
        # Get the class to use:
        class_name = MAPPINGS.get(key, None)
        if not class_name:
            continue
        base_list.append(class_name.from_dict(i))

    return len(base_list), base_list


def handle_record_attrs(
    dataclass, subkey: Literal["form", "server", "comments"]
) -> None:
    """
    Sets the sslOnly and authVault attributes
    on a record. Backend function.

    Args:
        dataclass: The dataclass to set the attributes on
        subkey: The subkey to set the attributes on

    Returns:
        None
    """

    if subkey not in ["formRecord", "serverRecord", "Comment"]:
        raise ValueError("subkey must be one of 'form', 'server', or 'Comment'")
    if subkey != "Comment":
        setattr(
            dataclass,
            f"{subkey}_authVault",
            getattr(dataclass, subkey).get("authVault") == "true",
        )
        setattr(
            dataclass,
            f"{subkey}_sslOnly",
            getattr(dataclass, subkey).get("sslOnly") == "true",
        )
        setattr(dataclass, f"{subkey}_type", getattr(dataclass, subkey).get("type"))
        if "seleniumCreds" in getattr(dataclass, subkey):
            setattr(
                dataclass,
                f"{subkey}_seleniumCreds",
                getattr(dataclass, subkey).get("seleniumCreds") == "true",
            )

        list_data = handle_qualys_list(getattr(dataclass, subkey), subkey)
        setattr(dataclass, f"{subkey}_fields_count", list_data[0])
        setattr(dataclass, f"{subkey}_fields_list", list_data[1])
        setattr(dataclass, subkey, None)
    else:
        list_data = handle_qualys_list(getattr(dataclass, "comments"), subkey)
        setattr(dataclass, f"comments_count", list_data[0])
        setattr(dataclass, f"comments_list", list_data[1])
        setattr(dataclass, "comments", None)


def handle_date_by_attrs(dataclass, subkey: Literal["createdBy", "updatedBy"]) -> None:
    """
    Handles date fields by attributes

    Args:
        dataclass: The dataclass to handle
        subkey: The subkey to handle

    Returns:
        None
    """
    if subkey not in ["createdBy", "updatedBy", "owner"]:
        raise ValueError("subkey must be one of 'createdBy', 'updatedBy', 'owner'")

    if getattr(dataclass, subkey):
        setattr(dataclass, f"{subkey}_id", int(getattr(dataclass, subkey).get("id")))
        setattr(
            dataclass, f"{subkey}_username", getattr(dataclass, subkey).get("username")
        )
        setattr(
            dataclass,
            f"{subkey}_firstName",
            getattr(dataclass, subkey).get("firstName"),
        )
        setattr(
            dataclass, f"{subkey}_lastName", getattr(dataclass, subkey).get("lastName")
        )
        setattr(dataclass, subkey, None)


@dataclass
class WebAppAuthRecord:
    """
    Represents an authentication record
    in Qualys WAS
    """

    id: Union[str, int] = None
    name: str = None
    owner: None = None
    # owner is parsed into below fields:
    owner_id: Union[str, int] = None
    owner_username: str = None
    owner_firstName: str = None
    owner_lastName: str = None
    # end owner
    formRecord: None = None
    # form_record is parsed into below fields:
    formRecord_type: str = None
    formRecord_sslOnly: bool = None
    formRecord_authVault: bool = None
    formRecord_seleniumCreds: bool = None
    formRecord_fields_count: Union[str, int] = 0
    formRecord_fields_list: BaseList = None
    # end form_record
    serverRecord: None = None
    # serverRecord is parsed into below fields:
    serverRecord_type: str = None
    serverRecord_sslOnly: bool = None
    serverRecord_authVault: bool = None
    serverRecord_fields_count: Union[str, int] = 0
    serverRecord_fields_list: BaseList = None
    # end serverRecord
    oauth2Record: None = None
    # oauth2Record is parsed into below fields:
    oauth2Record_grantType: str = None
    oauth2Record_accessTokenUrl: str = None
    oauth2Record_clientId: str = None
    oauth2Record_clientSecret: str = None
    oauth2Record_scope: str = None
    oauth2Record_seleniumCreds: bool = None
    # end oauth2Record
    tags: None = None
    # tags is parsed into below fields:
    tags_count: Union[str, int] = 0
    tags_list: BaseList = None
    # end tags
    comments: None = None
    # comments is parsed into below fields:
    comments_count: Union[str, int] = 0
    comments_list: BaseList = None
    # end comments
    createdDate: Union[str, datetime] = None
    updatedDate: Union[str, datetime] = None
    createdBy: None = None
    # createdBy is parsed into below fields:
    createdBy_id: Union[str, int] = None
    createdBy_username: str = None
    createdBy_firstName: str = None
    createdBy_lastName: str = None
    # end createdBy
    updatedBy: None = None
    # updatedBy is parsed into below fields:
    updatedBy_id: Union[str, int] = None
    updatedBy_username: str = None
    updatedBy_firstName: str = None
    updatedBy_lastName: str = None
    # end updatedBy

    def __post_init__(self):
        setattr(self, "id", int(self.id))

        DT_FIELDS = ["createdDate", "updatedDate"]
        DATEBY_FIELDS = ["owner", "createdBy", "updatedBy"]
        RECORD_FIELDS = ["formRecord", "serverRecord", "comments"]

        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        for field in DATEBY_FIELDS:
            if getattr(self, field):
                handle_date_by_attrs(self, field)

        if self.tags:
            res = handle_qualys_list(self.tags, "Tag")
            setattr(self, "tags_count", res[0])
            setattr(self, "tags_list", res[1])
            setattr(self, "tags", None)

        for field in RECORD_FIELDS:
            if getattr(self, field):
                handle_record_attrs(self, field if field != "comments" else "Comment")

        if self.oauth2Record:
            build_oauth2_record(self)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def __str__(self):
        return str(self.name)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)



"""
Contains code to ensure XML kwargs are parsed correctly
"""

from .filter_mappings import ENDPOINT_MAPPINGS, FILTER_MAPPING


def validate_kwargs(endpoint: str, **kwargs):
    """
    Ensures that the kwargs provided are valid for the given endpoint.
    """

    # Validate endpoint:
    if endpoint not in ENDPOINT_MAPPINGS:
        raise ValueError(
            f"Invalid endpoint: {endpoint}. Acceptable endpoints: {', '.join(ENDPOINT_MAPPINGS.keys())}"
        )

    # If the lastScan_status kwarg is provided, or if the key contains '_operator',
    # .upper() the value to ensure it's in the correct format:
    for key in kwargs.keys():
        if "lastScan_status" in key or "_operator" in key:
            kwargs[key] = str(kwargs[key]).upper()

    # Get the endpoint's filter data types.
    # ENDPOINT_MAPPINGS is a dictionary that for each endpoint,
    # contains the Qualys-expected data types for each field.
    # Combining this with FILTER_MAPPING, which contains the
    # operators available for each data type, we can validate
    # the provided kwargs.
    endpoint_dtypes = ENDPOINT_MAPPINGS[endpoint]

    # If a _operator key is provided, ensure that the
    # married kwarg is also provided:
    for key in [i for i in kwargs.keys() if "_operator" in i]:
        if key[:-9] not in kwargs:
            raise ValueError(f"Missing filter kwarg for {key[:-9]}")

    # Check if the provided kwargs are valid
    for key, value in kwargs.items():
        if "_operator" not in key and key not in endpoint_dtypes:
            raise ValueError(
                f"Invalid key: {key}. Acceptable keys: {endpoint_dtypes.keys()}"
            )

        if key.endswith("_operator"):
            if value not in FILTER_MAPPING[endpoint_dtypes[key[:-9]]]:
                raise ValueError(
                    f"Invalid operator for {key[:-9]}: {value}. Valid operators for {key[:-9]}: {FILTER_MAPPING[endpoint_dtypes[key[:-9]]]}"
                )

    # Lastly, convert any _s to .s in the keys:
    # Yikes...
    return {
        key.replace("_", ".") if key != "_operator" else key: value
        for key, value in kwargs.items()
    }



"""
Contains data classes for the WAS module.
"""

from .WebApp import WebApp
from .Tag import WASTag
from .AuthRecord import AuthRecord
from .CrawlingScript import CrawlingScript
from .PostmanCollection import PostmanCollection
from .SwaggerFile import SwaggerFile
from .UrlEntry import UrlEntry
from .Comment import Comment
from .WebAppAuthRecord import WebAppAuthRecord
from .WebAppAuthFormRecord import WebAppAuthFormRecord
from .WebAppAuthServerRecord import WebAppAuthServerRecord



"""
Contains the BaseList class for the qualysdk package.

The BaseList class is used to contain custom class objects in a list.
"""


class BaseList(list):
    """
    BaseList - represents a base list class for the qualysdk package.

    Essentially, this is a regular Python list with some additional functionality.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        # instead of returning "[...]", return a comma-separated string of the objects in the list
        return ", ".join(str(obj) for obj in self)



"""
Contains the RemediationActivity dataclass.
"""

from dataclasses import dataclass, asdict
from typing import Union, Literal
from datetime import datetime

from ...base.base_list import BaseList
from ...exceptions.Exceptions import *


@dataclass
class RemediationActivity:
    """
    Represents a remediation activity in TotalCloud.
    """

    resourceId: str = None
    controlId: int = None
    cloudType: Literal["AWS", "AZURE"] = None
    # accountId depends on the cloudType
    accountId: Union[str, int] = None
    region: str = None
    status: str = None
    resourceType: str = None
    remediationAction: str = None
    connectorName: str = None
    policyNames: BaseList[str] = None
    controlName: str = None
    triggeredOn: Union[str, datetime] = None
    Errors: str = None
    triggeredBy: str = None
    remediationReason: str = None

    def __post_init__(self):
        DT_FIELDS = ["triggeredOn"]

        for field in DT_FIELDS:
            if not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        if self.cloudType:
            if self.cloudType.upper() == "AWS":
                self.accountId = int(self.accountId)

        if self.Errors:
            setattr(self, "errors", str(self.Errors))

        if self.policyNames:
            bl = BaseList()
            for policy in self.policyNames:
                bl.append(policy)
            setattr(self, "policyNames", bl)

    def to_dict(self) -> dict:
        """
        Converts the object to a dictionary.

        Returns:
            dict: The object as a dictionary.
        """
        return asdict(self)

    def __dict__(self) -> dict:
        return self.to_dict()

    def keys(self) -> list:
        """
        Returns the keys of the object.

        Returns:
            list: The keys of the object.
        """
        return self.to_dict().keys()

    def values(self) -> list:
        """
        Returns the values of the object.

        Returns:
            list: The values of the object.
        """
        return self.to_dict().values()

    def items(self) -> list:
        """
        Returns the items of the object.

        Returns:
            list: The items of the object.
        """
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        Creates a RemediationActivity object from a dictionary.

        Args:
            data (dict): The dictionary to create the object from.

        Returns:
            RemediationActivity: The object created from the dictionary.
        """
        return cls(**data)



"""
SQL server integration for qualysdk
"""

from .base import db_connect
from .vmdr import (
    upload_vmdr_ags,
    upload_vmdr_kb,
    upload_vmdr_hosts,
    upload_vmdr_ips,
    upload_vmdr_hld,
    upload_vmdr_scanners,
    upload_vmdr_static_search_lists,
    upload_vmdr_dynamic_search_lists,
    upload_vmdr_users,
    upload_vmdr_scan_list,
    upload_vmdr_report_list,
    upload_vmdr_scheduled_report_list,
    upload_vmdr_template_list,
    upload_vmdr_kb_qvs,
    upload_vmdr_activity_log,
)

from .gav import upload_gav_hosts

from .totalcloud import (
    upload_totalcloud_aws_connectors,
    upload_totalcloud_azure_connectors,
    upload_totalcloud_gcp_connectors,
    upload_totalcloud_control_metadata,
    upload_totalcloud_aws_ec2,
    upload_totalcloud_aws_s3,
    upload_totalcloud_aws_acl,
    upload_totalcloud_aws_rds,
    upload_totalcloud_aws_iamuser,
    upload_totalcloud_aws_vpc,
    upload_totalcloud_aws_securitygroup,
    upload_totalcloud_aws_lambda,
    upload_totalcloud_aws_subnet,
    upload_totalcloud_aws_internetgateway,
    upload_totalcloud_aws_loadbalancer,
    upload_totalcloud_aws_routetable,
    upload_totalcloud_aws_ebsvolume,
    upload_totalcloud_aws_autoscalinggroup,
    upload_totalcloud_aws_ekscluster,
    upload_totalcloud_aws_eksnodegroup,
    upload_totalcloud_aws_fargateprofile,
    upload_totalcloud_aws_vpcendpoint,
    upload_totalcloud_aws_vpcendpointservice,
    upload_totalcloud_aws_iamgroup,
    upload_totalcloud_aws_iampolicy,
    upload_totalcloud_aws_iamrole,
    upload_totalcloud_aws_sagemakernotebook,
    upload_totalcloud_aws_cloudfrontdistribution,
    upload_totalcloud_remediation_activities,
)

from .cloud_agent import upload_cloud_agents
from .cs import upload_cs_containers
from .was import upload_was_webapps, upload_was_authentication_records



"""
Evaluation data class - contains data on a cloud provider account's standing with a control.
"""

from typing import Literal, Union
from dataclasses import dataclass, asdict
from datetime import datetime

from ...base.base_list import BaseList


@dataclass
class Evaluation:
    """
    Evaluation statistics for a control on a single resource ID,
    on a single cloud provider account.
    """

    firstEvaluated: Union[str, datetime] = None
    lastEvaluated: Union[str, datetime] = None
    dateReopen: Union[str, datetime] = None
    dateFixed: Union[str, datetime] = None

    def __post_init__(self):
        fields = self.__dataclass_fields__.keys()

        for field in fields:
            value = getattr(self, field)
            if not isinstance(value, datetime):
                try:
                    setattr(self, field, datetime.fromisoformat(value))
                except:
                    setattr(self, field, None)

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @staticmethod
    def from_dict(data):
        return Evaluation(**data)


@dataclass
class AccountLevelEvaluation:
    """
    Represents the evaluation of a control on a cloud provider account,
    across all resources under that account/control.
    """

    resourceId: str = None
    region: str = None
    accountId: str = None
    evaluatedOn: Union[str, datetime] = None
    evidences: list[dict[str, str]] = None
    resourceType: str = None
    connectorId: str = None
    result: Literal["PASS", "FAIL"] = None
    evaluationDates: Evaluation = None

    def __post_init__(self):
        if getattr(self, "evaluatedOn") and not isinstance(
            getattr(self, "evaluatedOn"), datetime
        ):
            setattr(
                self,
                "evaluatedOn",
                datetime.fromisoformat(getattr(self, "evaluatedOn")),
            )

        if self.evaluationDates:
            data = self.evaluationDates
            setattr(self, "evaluationDates", Evaluation.from_dict(data))

        if self.evidences:
            bl = BaseList()
            for evidence in self.evidences:
                bl.append(evidence)
            setattr(self, "evidences", bl)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create an AccountLevelEvaluation object from a dictionary
        """
        return cls(**data)



"""
assetgroups.py - AG manipulation functions for the Qualys VMDR module.
"""

from typing import *
from urllib.parse import parse_qs, urlparse
from ipaddress import IPv4Address, IPv6Address

from ..auth import BasicAuth
from .data_classes import AssetGroup
from ..base import *


def get_ag_list(
    auth: BasicAuth, page_count: Union["all", int] = "all", **kwargs
) -> list[AssetGroup]:
    """
    Gets a list of asset groups from the Qualys subscription.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.
        page_count (Union["all", int]): The number of pages to retrieve. Defaults to "all". If an integer is passed, that number of pages will be retrieved.

    :Kwargs:
        ```
        action (str): Action to perform on the asset groups. Defaults to "list". WARNING: SDK automatically sets this value to list. It is just included for completeness.
        echo_request (bool): Whether to echo the request. Defaults to False. WARNING: SDK automatically sets this value to 0. It is just included for completeness.
        output_format (Literal["csv", "xml"]): The output format of the response. Defaults to "xml". WARNING: SDK automatically sets this value to xml. It is just included for completeness.
        ids (str): The ID of the asset group to get. Defaults to None (all), but can be a single ID or a comma-separated string of IDs.
        id_min (int): The minimum ID of the asset group to get. Defaults to None.
        id_max (int): The maximum ID of the asset group to get. Defaults to None.
        truncation_limit (int): The truncation limit of the asset groups. Defaults to all records. 0 indicates all records. If non-0, the response will be truncated to the specified number of records. Pagination is handled automatically.
        network_ids (str): The network IDs of the asset groups to get. Defaults to None. Can be a single ID or a comma-separated string of IDs. WARNING: This has to be enabled in the Qualys subscription!
        unit_id (str): The unit ID of the asset groups to get. Defaults to None. Must be a single ID.
        user_id (str): The user ID of the asset groups to get. Defaults to None. Must be a single ID.
        title (str): The title of the asset groups to get. Defaults to None. Must be an exact match.
        show_attributes (Union[None, str]): Choose what attributes are returned. Defaults to None (show basic attrs), can be "ALL", "ID", "TITLE", .. For full list, see Qualys documentation: https://cdn2.qualys.com/docs/qualys-api-vmpc-user-guide.pdf
        ```
    Returns:
        BaseList[AssetGroup]: BaseList object containing the AssetGroup objects.
    """

    if type(page_count) in [int, float] and page_count <= 0:
        raise ValueError("page_count must be 'all' or an integer greater than 0.")

    results = BaseList()
    pulled = 0

    while True:
        kwargs["action"] = "list"
        kwargs["echo_request"] = False
        kwargs["output_format"] = "xml"

        # Enforce uppercase for show_attributes:
        if kwargs.get("show_attributes"):
            kwargs["show_attributes"] = kwargs["show_attributes"].upper()

        response = call_api(
            auth=auth,
            module="vmdr",
            endpoint="get_ag_list",
            params=kwargs,
            headers={"X-Requested-With": "qualysdk SDK"},
        )

        if response.status_code == 200:
            data = xml_parser(response.text)["ASSET_GROUP_LIST_OUTPUT"]

            if "ASSET_GROUP" not in data["RESPONSE"]["ASSET_GROUP_LIST"]:
                print("No asset groups found. Returning empty BaseList.")
                break

            # Check if type(data["RESPONSE"]["ASSET_GROUP_LIST"]["ASSET_GROUP"]) is dict.
            # If so, put it inside a list to normalize the class instantiation.
            if isinstance(data["RESPONSE"]["ASSET_GROUP_LIST"]["ASSET_GROUP"], dict):
                data["RESPONSE"]["ASSET_GROUP_LIST"]["ASSET_GROUP"] = [
                    data["RESPONSE"]["ASSET_GROUP_LIST"]["ASSET_GROUP"]
                ]

            for ag in data["RESPONSE"]["ASSET_GROUP_LIST"]["ASSET_GROUP"]:
                results.append(AssetGroup(**ag))

            pulled += 1
            # Check page count:
            if page_count != "all" and pulled >= page_count:
                print(f"Page count reached. Returning {pulled} pages.")
                break

            # Check for pagination:
            if data["RESPONSE"].get("WARNING"):
                # Get the id_min param:
                url = data["RESPONSE"]["WARNING"]["URL"]
                parsed_url = urlparse(url)
                id_min = parse_qs(parsed_url.query)["id_min"][0]
                kwargs["id_min"] = id_min
                print(f"Pagination detected. new id_min param: {id_min}")
            else:
                break

    return results


def manage_ag(
    auth: BasicAuth,
    action: Literal["add", "edit", "delete"],
    id: Union[AssetGroup, BaseList, str] = None,
    **kwargs,
) -> str:
    """
    Main function to perform an action on an asset group.

    Gets called by the add_ag, edit_ag, and delete_ag functions.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.
        action (Literal["add", "edit", "delete"]): The action to perform on the asset group.
        id (Union[AssetGroup, BaseList, str]): The ID of the asset group to edit or delete. Defaults to None. If a single AssetGroup, or a BaseList of AssetGroups is passed, the function will iterate over them.

    :Kwargs:
        ```
        echo_request (bool): Whether to echo the request. Defaults to False. WARNING: SDK automatically sets this value to 0. It is just included for completeness.

        (WHEN action=="add"):
            title (str): Title of the asset group. Required.
            comments (str): Comments for the asset group.
            division (str): Division of the asset group.
            function (str): Function of the asset group.
            location (str): Location of the asset group.
            business_impact (Literal["critical", "high", "medium", "low", "none"]): Business impact of the asset group.
            ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to add to the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
            appliance_ids (Union[BaseList[int], str]): Appliance IDs to add to the asset group. Comma-separated string, or a BaseList of integers.
            default_appliance_id (int): Default appliance ID for the asset group.
            domains (Union[BaseList[str], str]): Domains to add to the asset group. Comma-separated string, or a BaseList of strings.
            dns_names (Union[BaseList[str], str]): DNS names to add to the asset group. Comma-separated string, or a BaseList of strings.
            netbios_names (Union[BaseList[str], str]): NetBIOS names to add to the asset group. Comma-separated string, or a BaseList of strings.
            cvss_enviro_cdp (Literal["high", "medium-high", "low-medium", "low", "none"]): CVSS environmental CDP for the asset group.
            cvss_enviro_td (Literal["high", "medium", "low", "none"]): CVSS environmental TD for the asset group.
            cvss_enviro_cr (Literal["high", "medium", "low"]): CVSS environmental CR for the asset group.
            cvss_enviro_ir (Literal["high", "medium", "low"]): CVSS environmental IR for the asset group.
            cvss_enviro_ar (Literal["high", "medium", "low"]): CVSS environmental AR for the asset group.

        (WHEN action=="edit"):
            set_comments (str): New comments for the asset group.
            set_division (str): New division for the asset group.
            set_function (str): New function for the asset group.
            set_location (str): New location for the asset group.
            set_business_impact (Literal["critical", "high", "medium", "low", "none"]): New business impact for the asset group.
            add_ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to add to the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
            remove_ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to remove from the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
            set_ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to set for the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
            add_appliance_ids (Union[BaseList[int], str]): Appliance IDs to add to the asset group. Comma-separated string, or a BaseList of integers.
            remove_appliance_ids (Union[BaseList[int], str]): Appliance IDs to remove from the asset group. Comma-separated string, or a BaseList of integers.
            set_appliance_ids (Union[BaseList[int], str]): Appliance IDs to set for the asset group. Comma-separated string, or a BaseList of integers.
            set_default_appliance_id (int): New default appliance ID for the asset group.
            add_domains (Union[BaseList[str], str]): Domains to add to the asset group. Comma-separated string, or a BaseList of strings.
            remove_domains (Union[BaseList[str], str]): Domains to remove from the asset group. Comma-separated string, or a BaseList of strings.
            set_domains (Union[BaseList[str], str]): Domains to set for the asset group. Comma-separated string, or a BaseList of strings.
            add_dns_names (Union[BaseList[str], str]): DNS names to add to the asset group. Comma-separated string, or a BaseList of strings.
            remove_dns_names (Union[BaseList[str], str]): DNS names to remove from the asset group. Comma-separated string, or a BaseList of strings.
            set_dns_names (Union[BaseList[str], str]): DNS names to set for the asset group. Comma-separated string, or a BaseList of strings.
            add_netbios_names (Union[BaseList[str], str]): NetBIOS names to add to the asset group. Comma-separated string, or a BaseList of strings.
            remove_netbios_names (Union[BaseList[str], str]): NetBIOS names to remove from the asset group. Comma-separated string, or a BaseList of strings.
            set_netbios_names (Union[BaseList[str], str]): NetBIOS names to set for the asset group. Comma-separated string, or a BaseList of strings.
            set_title (str): New title for the asset group.
            set_cvss_enviro_cdp (Literal["high", "medium-high", "low-medium", "low", "none"]): New CVSS environmental CDP for the asset group.
            set_cvss_enviro_td (Literal["high", "medium", "low", "none"]): New CVSS environmental TD for the asset group.
            set_cvss_enviro_cr (Literal["high", "medium", "low"]): New CVSS environmental CR for the asset group.
            set_cvss_enviro_ir (Literal["high", "medium", "low"]): New CVSS environmental IR for the asset group.
            set_cvss_enviro_ar (Literal["high", "medium", "low"]): New CVSS environmental AR for the asset group.

        (WHEN action=="delete"):
            A single value in the id parameter.
        ```
    """

    # First, check the action:
    if action not in ["add", "edit", "delete"]:
        raise ValueError("action must be 'add', 'edit', or 'delete'.")

    kwargs["action"] = action

    POSSIBLE_LIST_ARGS = [
        "id",
        "ids",
        "ips",
        "appliance_ids",
        "domains",
        "dns_names",
        "netbios_names",
        "add_ips",
        "remove_ips",
        "set_ips",
        "add_appliance_ids",
        "remove_appliance_ids",
        "set_appliance_ids",
        "add_domains",
        "remove_domains",
        "set_domains",
        "add_dns_names",
        "remove_dns_names",
        "set_dns_names",
        "add_netbios_names",
        "remove_netbios_names",
        "set_netbios_names",
    ]
    POSSIBLE_IP_OBJ_ARGS = ["ips", "add_ips", "remove_ips", "set_ips"]

    # Look for single IP objects and convert them to strings:
    for arg in POSSIBLE_IP_OBJ_ARGS:
        if arg in kwargs and any(
            isinstance(kwargs[arg], obj) for obj in [IPv4Address, IPv6Address]
        ):
            kwargs[arg] = str(kwargs[arg])

    # Check if any list args are passed.
    # If so, convert them to comma-separated strings:
    for arg in POSSIBLE_LIST_ARGS:
        if arg in kwargs:
            if isinstance(kwargs[arg], BaseList):
                kwargs[arg] = [str(item) for item in kwargs[arg]]
                kwargs[arg] = ",".join(kwargs[arg])

    # Check if id is a single AssetGroup object:
    if isinstance(id, AssetGroup):
        id = id.id

    # Check if id is a BaseList object:
    if isinstance(id, BaseList):
        id = [str(ag.id) for ag in id]
        id = ",".join(id)

    if id:
        kwargs["id"] = id

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="manage_ag",
        payload=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    result = xml_parser(response.text)["SIMPLE_RETURN"]["RESPONSE"]["TEXT"]

    return result


def add_ag(auth: BasicAuth, title: str, **kwargs) -> str:
    """
    Adds an asset group to the Qualys subscription.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.
        title (str): Title of the asset group.

    :Kwargs:
        ```
        comments (str): Comments for the asset group.
        division (str): Division of the asset group.
        function (str): Function of the asset group.
        location (str): Location of the asset group.
        business_impact (Literal["critical", "high", "medium", "low", "none"]): Business impact of the asset group.
        ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to add to the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
        appliance_ids (Union[BaseList[int], str]): Appliance IDs to add to the asset group. Comma-separated string, or a BaseList of integers.
        default_appliance_id (int): Default appliance ID for the asset group.
        domains (Union[BaseList[str], str]): Domains to add to the asset group. Comma-separated string, or a BaseList of strings.
        dns_names (Union[BaseList[str], str]): DNS names to add to the asset group. Comma-separated string, or a BaseList of strings.
        netbios_names (Union[BaseList[str], str]): NetBIOS names to add to the asset group. Comma-separated string, or a BaseList of strings.
        cvss_enviro_cdp (Literal["high", "medium-high", "low-medium", "low", "none"]): CVSS environmental CDP for the asset group.
        cvss_enviro_td (Literal["high", "medium", "low", "none"]): CVSS environmental TD for the asset group.
        cvss_enviro_cr (Literal["high", "medium", "low"]): CVSS environmental CR for the asset group.
        cvss_enviro_ir (Literal["high", "medium", "low"]): CVSS environmental IR for the asset group.
        cvss_enviro_ar (Literal["high", "medium", "low"]): CVSS environmental AR for the asset group.
        ```
    Returns:
        str: Response text from the Qualys API.
    """

    return manage_ag(auth, action="add", title=title, **kwargs)


def edit_ag(auth: BasicAuth, id: Union[AssetGroup, BaseList, str], **kwargs) -> str:
    """
    Edits an asset group in the Qualys subscription.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.
        id (Union[AssetGroup, BaseList, str]): The ID of the asset group to edit. If a single AssetGroup, or a BaseList of AssetGroups is passed, the function will iterate over them.

    :Kwargs:
        ```
        echo_request (bool): Whether to echo the request. Defaults to False. WARNING: SDK automatically sets this value to 0. It is just included for completeness.
        set_comments (str): New comments for the asset group.
        set_division (str): New division for the asset group.
        set_function (str): New function for the asset group.
        set_location (str): New location for the asset group.
        set_business_impact (Literal["critical", "high", "medium", "low", "none"]): New business impact for the asset group.
        add_ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to add to the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
        remove_ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to remove from the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
        set_ips (Union[str, ipaddress.IPVxAddress, BaseList[ipaddress.IPVxAddress]]): IP addresses to set for the asset group. Comma-separated string, a single ipaddress.IPvxAddress, or a BaseList of ipaddress.IPvxAddress.
        add_appliance_ids (Union[BaseList[int], str]): Appliance IDs to add to the asset group. Comma-separated string, or a BaseList of integers.
        remove_appliance_ids (Union[BaseList[int], str]): Appliance IDs to remove from the asset group. Comma-separated string, or a BaseList of integers.
        set_appliance_ids (Union[BaseList[int], str]): Appliance IDs to set for the asset group. Comma-separated string, or a BaseList of integers.
        set_default_appliance_id (int): New default appliance ID for the asset group.
        add_domains (Union[BaseList[str], str]): Domains to add to the asset group. Comma-separated string, or a BaseList of strings.
        remove_domains (Union[BaseList[str], str]): Domains to remove from the asset group. Comma-separated string, or a BaseList of strings.
        set_domains (Union[BaseList[str], str]): Domains to set for the asset group. Comma-separated string, or a BaseList of strings.
        add_dns_names (Union[BaseList[str], str]): DNS names to add to the asset group. Comma-separated string, or a BaseList of strings.
        remove_dns_names (Union[BaseList[str], str]): DNS names to remove from the asset group. Comma-separated string, or a BaseList of strings.
        set_dns_names (Union[BaseList[str], str]): DNS names to set for the asset group. Comma-separated string, or a BaseList of strings.
        add_netbios_names (Union[BaseList[str], str]): NetBIOS names to add to the asset group. Comma-separated string, or a BaseList of strings.
        remove_netbios_names (Union[BaseList[str], str]): NetBIOS names to remove from the asset group. Comma-separated string, or a BaseList of strings.
        set_netbios_names (Union[BaseList[str], str]): NetBIOS names to set for the asset group. Comma-separated string, or a BaseList of strings.
        set_title (str): New title for the asset group.
        set_cvss_enviro_cdp (Literal["high", "medium-high", "low-medium", "low", "none"]): New CVSS environmental CDP for the asset group.
        set_cvss_enviro_td (Literal["high", "medium", "low", "none"]): New CVSS environmental TD for the asset group.
        set_cvss_enviro_cr (Literal["high", "medium", "low"]): New CVSS environmental CR for the asset group.
        set_cvss_enviro_ir (Literal["high", "medium", "low"]): New CVSS environmental IR for the asset group.
        set_cvss_enviro_ar (Literal["high", "medium", "low"]): New CVSS environmental AR for the asset group.
        ```
    Returns:
        str: Response text from the Qualys API.
    """

    return manage_ag(auth, action="edit", id=id, **kwargs)


def delete_ag(auth: BasicAuth, id: Union[AssetGroup, str]) -> str:
    """
    Deletes a single asset group from the Qualys subscription.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.
        id (Union[AssetGroup, str]): The ID of the asset group to delete.

    Returns:
        str: Response text from the Qualys API.
    """

    return manage_ag(auth, action="delete", id=id)



"""
Contains the WebAppAuthFormRecord class for the WAS
"""

from dataclasses import dataclass, asdict
from typing import Union


@dataclass
class WebAppAuthFormRecord:
    """
    Represents a single form authentication record
    in Qualys WAS
    """

    id: Union[str, int] = None
    name: str = None
    secured: bool = None
    value: str = None

    def __post_init__(self):
        if not isinstance(self.id, (str, int)):
            raise ValueError("id must be a string or integer")
        setattr(self, "id", int(self.id))

        setattr(self, "secured", self.secured == "true")

    def __str__(self):
        return f"{self.name} : {self.value}"

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def redact_password(self):
        """
        Obfuscates the password
        """
        if self.name.lower() == "password" or self.secured:
            setattr(self, "value", "REDACTED")



"""
contains the Control dataclass for the TotalCloud API
"""

from dataclasses import dataclass, asdict
from typing import Union, Literal
from datetime import datetime

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning

# suppress warning from bs4
import warnings

warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)

from ...base.base_list import BaseList


@dataclass
class Control:
    """
    Represents a single control that Qualys checks a cloud provider for.
    """

    cid: int = None
    controlName: str = None
    created: Union[str, datetime] = None
    modified: Union[str, datetime] = None
    controlType: str = None
    provider: str = None
    isCustomizable: bool = None
    serviceType: str = None
    criticality: str = None
    evaluation: None = None
    # BELOW FIELDS ARE PARSED OUT OF EVALUATION:
    evaluation_description: str = None  # NEEDS TO BE PARSED WITH BS4!
    evaluation_passMessage: str = None
    evaluation_failMessage: str = None
    evaluation_criteria: BaseList[str] = None  # come back to this one?
    # END OF EVALUATION FIELDS
    specification: str = None  # NEEDS TO BE PARSED WITH BS4!
    rationale: str = None  # NEEDS TO BE PARSED WITH BS4!
    manualRemediation: str = None  # NEEDS TO BE PARSED WITH BS4!
    references: str = None  # NEEDS TO BE PARSED WITH BS4!
    buildTimeRemediation: str = None  # NEEDS TO BE PARSED WITH BS4!
    resourceType: str = None
    remediationEnabled: bool = None
    policyNames: BaseList[str] = None
    executionType: str = None
    workflowBased: bool = None
    templateType: BaseList[str] = None

    def __post_init__(self):
        """
        __post_init__ - post initialization method for the Control dataclass
        """
        if self.created:
            setattr(self, "created", datetime.fromisoformat(self.created))

        if self.modified:
            setattr(self, "modified", datetime.fromisoformat(self.modified))

        if self.evaluation:
            FIELDS = ["description", "passMessage", "failMessage"]
            for field in FIELDS:
                if self.evaluation.get(field):
                    setattr(
                        self, f"evaluation_{field}", self.evaluation.get(field, None)
                    )
            if self.evaluation.get("criteria"):
                setattr(
                    self,
                    "evaluation_criteria",
                    BaseList(self.evaluation.get("criteria")),
                )
            self.evaluation = None

        BL_FIELDS = ["policyNames", "templateType"]
        for field in BL_FIELDS:
            if getattr(self, field):
                setattr(self, field, BaseList(getattr(self, field)))

        BS4_FIELDS = [
            "specification",
            "rationale",
            "manualRemediation",
            "references",
            "evaluation_description",
            "buildTimeRemediation",
        ]
        for field in BS4_FIELDS:
            if getattr(self, field):
                setattr(
                    self,
                    field,
                    BeautifulSoup(getattr(self, field), "html.parser").get_text(),
                )

    def to_dict(self):
        """
        to_dict - return the Control as a dictionary
        """
        return asdict(self)

    def __int__(self):
        return self.cid

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a Control object from a dictionary
        """
        return cls(**data)


@dataclass
class AccountLevelControl:
    """
    Represents control on the
    cloud provider account level.
    """

    controlName: str = None
    controlId: int = None
    policyNames: BaseList[str] = None
    criticality: str = None
    service: str = None
    result: str = None
    passedResources: int = None
    failedResources: int = None
    passWithExceptionResources: int = None

    def __post_init__(self):
        if getattr(self, "controlId") and not isinstance(
            getattr(self, "controlId"), int
        ):
            setattr(self, "controlId", int(getattr(self, "controlId")))

        if self.policyNames:
            data = self.policyNames
            bl = BaseList()
            if isinstance(data, dict):
                data = [data]
            for name in data:
                bl.append(name)

    def __int__(self):
        return self.controlId

    def __dict__(self):
        return self.to_dict()

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create an AccountLevelControl object from a dictionary
        """
        return cls(**data)



"""
Contains the WASTag class for WAS
"""

from dataclasses import dataclass, asdict


@dataclass
class WASTag:
    """
    Represents a single tag in WAS
    """

    id: int = None
    name: str = None

    def __post_init__(self):
        if self.id and isinstance(self.id, str):
            try:
                self.id = int(self.id)
            except ValueError:
                raise ValueError("id must be numeric")

    def __str__(self):
        return self.name

    def __int__(self):
        return self.id

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @staticmethod
    def from_dict(data):
        return WASTag(**data)



"""
base.py - base authentication class for qualysdk
"""

import json
from dataclasses import dataclass, field
from typing import Optional, Literal, Self


from ..exceptions import (
    InvalidCredentialsError,
    InvalidTokenError,
    InvalidAuthTypeError,
)


@dataclass
class BaseAuthentication:
    """
    Base class for authentication with qualysdk.
    """

    username: str
    password: str = field(
        repr=False
    )  # Hide password from repr. If the user wants to see it that badly, they can do so manually with the password attribute.
    token: Optional[str] = field(default=None, repr=False)  # same goes for token ^^
    auth_type: Literal["basic", "token"] = field(init=False)

    def __post_init__(self) -> None:
        """
        Post-init method to determine auth_type based on if a token is passed or not.
        """
        if self.token is None:
            self.auth_type = "basic"
        else:
            self.auth_type = "token"

    def __str__(self) -> str:
        """
        String representation of the authentication object.
        """
        return f"Base authentication object for {self.username}"

    def validate_type(self):
        """
        Validate the authentication object.
        """
        if self.auth_type == "basic":
            if self.username is None or self.password is None:
                raise InvalidCredentialsError(
                    "Username and password must be provided for basic authentication."
                )
        elif self.auth_type == "token":
            if self.token is None:
                raise InvalidTokenError(
                    "Token must be provided for token authentication."
                )
        else:
            raise InvalidAuthTypeError(
                "Invalid authentication type. Must be 'basic' or 'token'."
            )

    def to_dict(self) -> dict:
        """
        Convert the authentication object to a dictionary.
        """
        return {
            "username": self.username,
            "password": self.password,
            "token": self.token,
            "auth_type": self.auth_type,
        }

    @classmethod
    def from_dict(cls, data: dict):
        """
        Create an authentication object from a dictionary.
        """
        return cls(**data)

    @classmethod
    def from_json_string(cls, data: str):
        """
        Create an authentication object from a JSON string.
        """
        return cls.from_dict(json.loads(data))

    def to_json_string(self, pretty: bool = False) -> str:
        """
        Convert the authentication object to a JSON string.

        Params:
        ```
        pretty: bool (default is False) - whether to pretty print the JSON string or not.
        ```
        """
        return json.dumps(self.to_dict(), indent=4 if pretty else None)

    def __eq__(self, other):
        """
        Equality comparison for authentication objects.
        """
        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """
        Inequality comparison for authentication objects.
        """
        return self.to_dict() != other.to_dict()



"""
cloud_agent.py - Contains the functions to upload supported cloud_agent API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame
from sqlalchemy import Connection, types
from sqlalchemy.dialects.mysql import TEXT

from .base import upload_data, prepare_dataclass
from ..base.base_list import BaseList


def upload_cloud_agents(
    hosts: BaseList,
    cnxn: Connection,
    table_name: str = "cloud_agent_agents",
    override_import_dt: datetime = None,
) -> int:
    """
    Upload results from ```cloud_agent.list_agents``` to a SQL database.

    Args:
        agents (BaseList): A BaseList of CloudAgent objects.
        cnxn (Connection): The Connection object to the SQL database.
        table_name (str): The name of the table to upload to. Defaults to "cloud_agent_agents".
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "id": types.Integer(),
        "name": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "created": types.DateTime(),
        "modified": types.DateTime(),
        "type": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "tags": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "sourceInfo": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # list
        "qwebHostId": types.Integer(),
        "lastComplianceScan": types.DateTime(),
        "lastSystemBoot": types.DateTime(),
        "lastLoggedOnUser": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "dnsHostName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_agentVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_agentId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_status": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_lastCheckedIn": types.DateTime(),
        "agentInfo_connectedFrom": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_location": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_locationGeoLatitude": types.Float(),
        "agentInfo_locationGeoLongitude": types.Float(),
        "agentInfo_chirpStatus": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_platform": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_activatedModule": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_manifestVersion": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_agentConfiguration_id": types.Integer(),
        "agentInfo_agentConfiguration_name": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_activationKey_activationId": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "agentInfo_activationKey_title": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "netbiosName": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "criticalityScore": types.Integer(),
        "lastVulnScan": types.DateTime(),
        "vulnsUpdated": types.DateTime(),
        "informationGatheredUpdated": types.DateTime(),
        "domain": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "fqdn": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "os": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "networkGuid": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "address": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "trackingMethod": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "manufacturer": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "model": types.String().with_variant(TEXT(charset="utf8"), "mysql", "mariadb"),
        "totalMemory": types.Integer(),
        "timezone": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "biosDescription": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
        "openPort": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "software": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "vuln": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "processor": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "volume": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "account": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "networkInterface": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # BaseList
        "isDockerHost": types.Boolean(),
        "dockerInfo": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),  # list
        "cloudProvider": types.String().with_variant(
            TEXT(charset="utf8"), "mysql", "mariadb"
        ),
    }

    # Prepare the dataclass for insertion:
    df = DataFrame([prepare_dataclass(host) for host in hosts])

    # Drop cols that are parsed out into other fields:
    df.drop(
        columns=[
            "agentInfo_activationKey",
            "agentInfo_agentConfiguration",
            "agentInfo",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(df, table_name, cnxn, COLS, override_import_dt)
